unit Main;

{$IMPORTEDDATA OFF}

interface

uses Windows, Messages;

const
  AppName = 'NkHotKey';
  WM_NKHOTKEY_NOTIFYICON_CALLBACK = WM_APP;
  WM_NKHOTKEY_COMMAND = WM_APP + 1;
  IDM_MENUITEM_START = 20000;

var
  MainWnd: HWND;
  ProfDir: WideString;
  KeymapFilePath: WideString;

function MakeWindow: HWND;
procedure MsgBox(const S: WideString; const Caption: WideString=AppName);
procedure ClearHotKeys;
procedure SuspendHotKeys;
procedure UnsuspendHotKeys;
procedure UpdateHotKeys(Reload: Boolean=True);

implementation

uses NTypes, NLib, NSys, NForm, NKeymap, NConf, NResource, Command, NotifyIcon,
  AboutDialog, KeymapDialog;

const
  IDM_SUSPEND = 10000;
  IDM_KEYMAP = 10001;
  IDM_RELOAD = 10002;
  IDM_ABOUT = 10003;
  IDM_EXIT = 10004;

var
  HotKeys: array of TStringArray;
  Keymap: TConf;
  TaskbarCreatedMsg: UINT;

function Decode(const StrArray: array of string): TWideStringArray;
var
  i: Integer;
begin
  SetLength(Result, Length(StrArray));
  for i := 0 to High(StrArray) do
    Result[i] := DecodeUTF8(StrArray[i]);
end;

function GetKeyNames(const Command: string): string;
var
  key: string;
begin
  Result := '';
  while Keymap.EachKeys(key) do
    if Keymap[key][0] = Command then
      if Result = '' then
        Result := StrToKeyName(key)
      else
        Result := Result + ', ' + StrToKeyName(key);
end;

procedure About;
var
  w: TWindow;
begin
  with TAboutDialog.Create do
    try
      ProfDir := Main.ProfDir;
      w := TWindow.Create;
      w.Handle := GetDesktopWindow;
      ShowModal(w);
      w.Free;
    finally
      Free;
    end;
end;

procedure ShowKeymap;
var
  i: Integer;
  buf: WideString;
  key: string;
  keys: array of string;
  w: TWindow;
begin
  SetLength(keys, Keymap.Count);
  i := 0;
  while Keymap.EachKeys(key) do
  begin
    keys[i] := key;
    Inc(i);
  end;
  Sort(keys);
  buf := '';
  for i := 0 to High(keys) do
    buf := buf + keys[i] + #9 + JoinCommandLine(Decode(Keymap[keys[i]])) + #13#10;

  with TKeymapDialog.Create do
    try
      Text := 'Keymap File: ' + KeymapFilePath + #13#10#13#10 + buf;
      w := TWindow.Create;
      w.Handle := GetDesktopWindow;
      ShowModal(w);
      w.Free;
    finally
      Free;
    end;
end;

procedure Reload;
begin
  UpdateHotKeys;
  SetNotifyIconInfo('Reload Keymap');
end;

procedure PopupMenu;
var
  pt: TPoint;
begin
  with TPopupMenu.Create do
    try
      AppendItem(IDM_SUSPEND, '&Suspend'#9 + GetKeyNames('suspend'));
      CheckItem(IDM_SUSPEND, Suspended);
      AppendItem(0, '-');
      AppendItem(IDM_KEYMAP, 'Show &Keymap');
      AppendItem(IDM_RELOAD, '&Reload Keymap');
      AppendItem(0, '-');
      AppendItem(IDM_ABOUT, '&About ' + AppName);
      AppendItem(0, '-');
      AppendItem(IDM_EXIT, '&Exit ' + AppName);
      SetForegroundWindow(MainWnd);
      GetCursorPos(pt);
      Popup(MainWnd, pt.x, pt.y);
    finally
      Free;
    end;
end;

function WindowProc(hWnd: HWND; Msg: UINT; wParam: WPARAM; lParam: LPARAM):
  LRESULT; stdcall;
var
  id: Word;
begin
  Result := 0;
  case Msg of
    WM_NKHOTKEY_NOTIFYICON_CALLBACK:
      if (lParam = WM_LBUTTONUP) or (lParam = WM_RBUTTONUP) then
        PopupMenu;
    WM_NKHOTKEY_COMMAND:
      DispatchCommand(Decode(HotKeys[wParam]));
    WM_HOTKEY:
      if (Length(HotKeys[wParam]) > 0) and (HotKeys[wParam][0] = 'menu') then
      begin
        EndMenu;
        PostMessageW(hWnd, WM_NKHOTKEY_COMMAND, wParam, 0);
      end else
        SendMessageW(hWnd, WM_NKHOTKEY_COMMAND, wParam, 0);
    WM_COMMAND:
      begin
        id := wParam and $ffff;
        case id of
          IDM_SUSPEND: DispatchCommand([WideString('suspend')]);
          IDM_KEYMAP: ShowKeymap;
          IDM_RELOAD: Reload;
          IDM_ABOUT: About;
          IDM_EXIT: PostQuitMessage(0);
        else
          if id >= IDM_MENUITEM_START then
            DispatchCommand(SliceArrayW(MenuItems[id-IDM_MENUITEM_START], 1));
        end;
      end;
  else
    if Msg = TaskbarCreatedMsg then
      ReAddNotifyIcon
    else
      Result := DefWindowProcW(hWnd, Msg, wParam, lParam);
  end;
end;

function MakeWindow: HWND;
const
  classname = AppName + 'WindowClass';
var
  wc: TWndClassW;
begin
  TaskbarCreatedMsg := RegisterWindowMessageW('TaskbarCreated');
  FillChar(wc, SizeOf(wc), 0);
  wc.lpfnWndProc := @WindowProc;
  wc.hInstance := HInstance;
  wc.lpszClassName := classname;
  if RegisterClassW(wc) = 0 then
    raise EOSError.Create;
  Result := CreateWindowExW(WS_EX_TOOLWINDOW , classname, nil, WS_POPUP, 0, 0,
    64, 64, 0, 0, HInstance, nil);
  if Result = 0 then
    raise EOSError.Create;
end;

procedure MsgBox(const S: WideString; const Caption: WideString=AppName);
begin
  MessageBoxW(0, PWideChar(S), PWideChar(Caption), MB_SETFOREGROUND);
end;

procedure ClearHotKeys;
var
  i: Integer;
begin
  for i := 0 to High(HotKeys) do
    UnregisterHotKey(MainWnd, i);
  Keymap.Clear;
end;

procedure SuspendHotKeys;
var
  i: Integer;
begin
  for i := 0 to High(HotKeys) do
    if HotKeys[i][0] <> 'suspend' then
      UnregisterHotKey(MainWnd, i);
end;

procedure UnsuspendHotKeys;
begin
  UpdateHotKeys(False);
end;

procedure UpdateHotKeys(Reload: Boolean=True);
var
  i: Integer;
  key, error: string;
  ks: TKeyState;
  m: UINT;
  vkey: Word;
begin
  if Reload then
  begin
    ClearHotKeys;
    Keymap.Read(ResourceText('KEYMAP'));
    Keymap.LoadFromFile(KeymapFilePath);
    SetLength(HotKeys, Keymap.Count);
  end else
    for i := 0 to High(HotKeys) do
      UnregisterHotKey(MainWnd, i);

  error := '';
  i := 0;
  while Keymap.EachKeys(key) do
  begin
    StrToVkey(key, ks, vkey);
    m := 0;
    if ksAlt in ks then
      m := m or MOD_ALT;
    if ksCtrl in ks then
      m := m or MOD_CONTROL;
    if ksShift in ks then
      m := m or MOD_SHIFT;
    if ksWin in ks then
      m := m or MOD_WIN;
    try
      if not RegisterHotKey(MainWnd, i, m, vkey) then
        raise EOSError.Create;
      HotKeys[i] := Keymap[key];
      Inc(i);
    except
      on E: EOSError do
        error := error + '  ' + key + ': ' + E.Msg + #10;
    end;
  end;
  if error <> '' then
    MsgBox('register hotkey failed:'#10 + error);
end;

initialization
  Keymap := TConf.Create;

finalization
  Keymap.Free;

end.
