unit NotifyIcon;

{$IMPORTEDDATA OFF}

interface

procedure AddNotifyIcon;
procedure DeleteNotifyIcon;
procedure ReAddNotifyIcon;
procedure SetNotifyIconState(Suspended: Boolean);
procedure SetNotifyIconInfo(const Info: WideString);

implementation

uses Windows, ShellAPI, NSys, Main;

const
  NIF_INFO = $10;

type
  // Shell32.dll Version >= 5.00 (Windows 2000 and later)
  TNotifyIconDataW = record
    cbSize: DWORD;
    Wnd: HWND;
    uID: UINT;
    uFlags: UINT;
    uCallbackMessage: UINT;
    hIcon: HICON;
    szTip: array [0..127] of WideChar;
    dwState: DWORD;
    dwStateMask: DWORD;
    szInfo: array [0..255] of WideChar;
    uTimeout: UINT;
    szInfoTitle: array [0..63] of WideChar;
    dwInfoFlags: DWORD;
  end;

var
  NotifyIconData: TNotifyIconDataW;
  AppIcon: HICON;
  SuspendedIcon: HICON;

procedure CopyWStr(const Src: WideString; var Dest: array of WideChar);
var
  len, maxlen: Integer;
begin
  len := Length(Src);
  maxlen := Length(Dest) - 1; // -1 for null terminate
  if len > maxlen then
    len := maxlen;
  Move(Src[1], Dest[0], len * SizeOf(WideChar));
  Dest[maxlen - 1] := WideChar(0);
end;

procedure AddNotifyIcon;
begin
  AppIcon := LoadImageW(HInstance, PWideChar(1), IMAGE_ICON, 16, 16,
      LR_DEFAULTCOLOR);
  SuspendedIcon := LoadImageW(HInstance, PWideChar(2), IMAGE_ICON, 16, 16,
      LR_DEFAULTCOLOR);
  NotifyIconData.cbSize := SizeOf(NotifyIconData);
  NotifyIconData.Wnd := MainWnd;
  NotifyIconData.uID := 1;
  NotifyIconData.uFlags := NIF_ICON or NIF_MESSAGE or NIF_TIP or NIF_INFO;
  NotifyIconData.uCallbackMessage := WM_NKHOTKEY_NOTIFYICON_CALLBACK;
  NotifyIconData.hIcon := AppIcon;
  CopyWStr(AppName, NotifyIconData.szTip);
  CopyWStr(AppName, NotifyIconData.szInfoTitle);
  if not Shell_NotifyIconW(NIM_ADD, @NotifyIconData) then
    raise EOSError.Create;
end;

procedure DeleteNotifyIcon;
begin
  if not Shell_NotifyIconW(NIM_DELETE, @NotifyIconData) then
    raise EOSError.Create;
end;

procedure ReAddNotifyIcon;
begin
  if not Shell_NotifyIconW(NIM_ADD, @NotifyIconData) then
    raise EOSError.Create;
end;

procedure SetNotifyIconState(Suspended: Boolean);
begin
  if Suspended then
  begin
    NotifyIconData.hIcon := SuspendedIcon;
    CopyWStr(AppName + #13#10'Suspended', NotifyIconData.szTip);
    CopyWStr('Suspended', NotifyIconData.szInfo);
  end
  else begin
    NotifyIconData.hIcon := AppIcon;
    CopyWStr(AppName, NotifyIconData.szTip);
    CopyWStr('Unsuspended', NotifyIconData.szInfo);
  end;
  if not Shell_NotifyIconW(NIM_MODIFY, @NotifyIconData) then
    raise EOSError.Create;
end;

procedure SetNotifyIconInfo(const Info: WideString);
begin
  CopyWStr(Info, NotifyIconData.szInfo);
  if not Shell_NotifyIconW(NIM_MODIFY, @NotifyIconData) then
    raise EOSError.Create;
end;

end.
