unit WinMan;

{$IMPORTEDDATA OFF}

interface

uses Windows;

type
  TWindowInfo = record
    Handle: HWND;
    Caption: WideString;
    ClassName: WideString;
    Left: Integer;
    Top: Integer;
    Width: Integer;
    Height: Integer;
    ErrorPosition: Boolean;
  end;

const
  spUndef = $7fffffff;

procedure Foreground(const ClassName: WideString; Reverse: Boolean);
function GetInfo: TWindowInfo;
procedure Minimize;
procedure MinimizeAll;
procedure SetPosition(L, T, R, B, W, H: Integer; Center: Boolean);
procedure ToggleMaximize;

implementation

uses Messages, WinAdd;

type
  TWndArray = array of HWND;

function GetClassName(Wnd: HWND): WideString;
begin
  SetString(Result, nil, 256);
  SetLength(Result, GetClassNameW(wnd, PWideChar(Result), Length(Result)));
end;

function ListWindow(const ClassName: WideString): TWndArray;
type
  TListWindowRec = record
    ClassName: WideString;
    Wnds: TWndArray;
    Count: Integer;
  end;
  PListWindowRec = ^TListWindowRec;

  function appendwnd(Wnd: HWND; lParam: LPARAM): BOOL; stdcall;
  begin
    with PListWindowRec(lParam)^ do begin
      Result := True;
      if not IsWindowVisible(Wnd)
          or (GetClassName(Wnd) <> ClassName)
          or (GetWindowLongW(Wnd, GWL_EXSTYLE) and WS_EX_TOPMOST <> 0) then
        Exit;
      if Count = Length(Wnds) then
        SetLength(Wnds, Count + Count div 4);
      Wnds[Count] := Wnd;
      Inc(Count);
    end;
  end;
var
  rec: TListWindowRec;
begin
  rec.ClassName := ClassName;
  SetLength(rec.Wnds, 256);
  rec.Count := 0;
  EnumWindows(@appendwnd, LPARAM(@rec));
  SetLength(rec.Wnds, rec.Count);
  Result := rec.Wnds;
end;

procedure Foreground(const ClassName: WideString; Reverse: Boolean);
var
  wnds: TWndArray;
  fg, insertafter: HWND;
begin
  wnds := ListWindow(ClassName);
  if Length(wnds) = 0 then
    Exit;

  insertafter := 0;
  if GetForegroundWindow = wnds[0] then begin
    if Length(wnds) = 1 then
      Exit;
    if Reverse then begin
      fg := wnds[High(wnds)];
    end else begin
      insertafter := wnds[High(wnds)];
      fg := wnds[1];
    end;
  end else
    fg := wnds[0];

  if IsIconic(fg) then
    PostMessageW(fg, WM_SYSCOMMAND, SC_RESTORE, 0);
  SetForegroundWindow(fg);
  if insertafter <> 0 then
    SetWindowPos(wnds[0], insertafter, 0, 0, 0, 0,
        SWP_NOSIZE or SWP_NOMOVE or SWP_NOACTIVATE or SWP_NOOWNERZORDER
        or SWP_ASYNCWINDOWPOS);
end;

function GetInfo: TWindowInfo;
var
  wp: TWindowPlacement;
begin
  with Result do begin
    Handle := GetForegroundWindow;
    ClassName := GetClassName(Handle);
    SetString(Caption, nil, 256);
    SetLength(Caption, GetWindowTextW(Handle, PWideChar(Caption), Length(Caption)));
    wp.length := SizeOf(wp);
    Result.ErrorPosition := not GetWindowPlacement(Handle, @wp);
    if not Result.ErrorPosition then begin
      Left := wp.rcNormalPosition.Left;
      Top := wp.rcNormalPosition.Top;
      Width := wp.rcNormalPosition.Right - Left;
      Height := wp.rcNormalPosition.Bottom - Top;
    end;
  end;
end;

procedure ToggleMaximize;
var
  wnd: HWND;
  ws: Integer;
begin
  wnd := GetForegroundWindow;
  ws := GetWindowLongW(wnd, GWL_STYLE);
  if (ws and WS_MAXIMIZEBOX) = 0 then
    Exit;
  if IsZoomed(wnd) then
    PostMessageW(wnd, WM_SYSCOMMAND, SC_RESTORE, 0)
  else
    // "PostMessage(wnd, WM_SYSCOMMAND, SC_MAXIMIZE, 0)"
    // to Win7 Explorer in F11 fullscreen mode
    ShowWindowAsync(wnd, SW_MAXIMIZE);
end;

procedure _Minimize(Wnd: HWND);
var
  rc: TRect;
  ws: Integer;
begin
  ws := GetWindowLongW(Wnd, GWL_STYLE);
  if ((ws and WS_VISIBLE) = 0) or
      ((ws and WS_MINIMIZE) <> 0) or
      ((ws and WS_MINIMIZEBOX) = 0) or
      ((GetWindowLongW(Wnd, GWL_EXSTYLE) and WS_EX_TOOLWINDOW) <> 0) or
      not GetWindowRect(Wnd, rc) or
      ((rc.Right-rc.Left) = 0) or ((rc.Bottom-rc.Top) = 0) then
    Exit;
  // ShowWindowAsync (SW_MINIMIZE) cannot minimize some windows
  PostMessageW(Wnd, WM_SYSCOMMAND, SC_MINIMIZE, 0);
end;

procedure Minimize;
begin
  _Minimize(GetForegroundWindow);
end;

procedure MinimizeAll;
  function enumwindowsproc(Wnd: HWND; lParam: LPARAM): BOOL; stdcall;
  begin
    _Minimize(Wnd);
    Result := True;
  end;
begin
  EnumWindows(@enumwindowsproc, 0);
end;

procedure SetPosition(L, T, R, B, W, H: Integer; Center: Boolean);
var
  wnd: HWND;
  rc, wrc: TRect;
  wp: TWindowPlacement;
begin
  wnd := GetForegroundWindow;
  if (GetWindowLongW(wnd, GWL_STYLE) and WS_THICKFRAME) = 0 then
    Exit;

  GetWindowRect(wnd, rc);
  SystemParametersInfo(SPI_GETWORKAREA, 0, @wrc, 0);

  if W = spUndef then
    W := rc.Right - rc.Left;
  if H = spUndef then
    H := rc.Bottom - rc.Top;

  if Center then begin
    L := wrc.Left + (wrc.Right -  wrc.Left - W) div 2;
    T := wrc.Top + (wrc.Bottom -  wrc.Top - H) div 2;
  end else begin
    if L = spUndef then
      if R = spUndef then
        L := rc.Left
      else
        L := wrc.Right - W - R;

    if T = spUndef then
      if B = spUndef then
        T := rc.Top
      else
        T := wrc.Bottom - H - B;
  end;

  wp.length := SizeOf(wp);
  wp.flags := WPF_ASYNCWINDOWPLACEMENT;
  if GetWindowPlacement(wnd, @wp) then begin
    with wp.rcNormalPosition do begin
      Left := L;
      Top := T;
      Right := L + W;
      Bottom := T + H;
    end;
    SetWindowPlacement(wnd, @wp);
  end;
end;

end.
